#include "laser_detect/laser_detect.h"


const double obj_dist = 0.5;
const int min_detect_dotsNum = 2;
const double max_two_dots_interval = 0.5;
const int min_dots_in_region = 2;
const double max_laser_valid_length = 10;

laser_detect::laser_detect()
{
   // boost::recursive_mutex::scoped_lock l(configuration_mutex_);
    ros::NodeHandle private_nh_("~");
    private_nh_.param("laser_frame_id", laser_frame_id_, std::string("laser"));
    private_nh_.param("base_frame_id", base_frame_id_, std::string("base_link"));
    private_nh_.param("global_frame_id", global_frame_id_, std::string("map"));
    private_nh_.param("detect_frame_id", detect_frame_id_, std::string("rival"));
    private_nh_.param("scan_topic", scan_topic_, std::string("scan"));

    requesStatictMap();


    tf_ = new tf::TransformListener();
    laser_detect_tf_ = new tf::TransformBroadcaster();

    laser_scan_sub_ = new message_filters::Subscriber<sensor_msgs::LaserScan>(nh_, scan_topic_, 5);

    laser_scan_filter_ = new tf::MessageFilter<sensor_msgs::LaserScan>(*laser_scan_sub_,*tf_, global_frame_id_, 5);
    laser_scan_filter_->registerCallback(boost::bind(&laser_detect::laserReceived,this, _1));


}


bool laser_detect::laser_scan_solve(int count,tf::Stamped<tf::Pose>& laserPose , UTILITY_REGION_Typedef& PublishPoint)
{
  obj_detect_dots.clear();
  obj_regions.clear();
  utility_obj_regions.clear();


  for(int i = 0; i < count; i++)
  {
      float hitLength = laser_scan_range[i][0];
      float hitAngle  = laser_scan_range[i][1];

      // Check for NaN
        if(hitLength != hitLength)
        {
          continue;
        }

        if(hitLength > max_laser_valid_length)
        {
          continue;
        }

        double pitch,roll,laserYaw;
        laserPose.getBasis().getEulerYPR(laserYaw, pitch, roll);

        double laserX = laserPose.getOrigin().x();
        double laserY = laserPose.getOrigin().y();

        double hitX = laserX + hitLength * cos(laserYaw + hitAngle);
        double hitY = laserY + hitLength * sin(laserYaw + hitAngle);

        //printf("laserX %f laserY %f hitX  %f   hitY   %f  hitLength %f,hitAngle %f laserYaw %f",laserX,laserY,hitX,hitY,hitLength,hitAngle*180/M_PI,laserYaw*180/M_PI);

        int map_x = MAP_GXWX(static_map_,hitX);
        int map_y = MAP_GYWY(static_map_,hitY);


        if(MAP_VALID(static_map_, map_x, map_y))
        {
            double obj_to_bound = static_map_->cells[MAP_INDEX(static_map_,map_x,map_y)].occ_dist;

            if(obj_to_bound > obj_dist)
            {
                LASER_DOT_Typedef laser_dots = {hitX,hitY,hitLength,hitAngle};
                obj_detect_dots.push_back(laser_dots);
            }
        }
  }

  int obj_detect_dotsNum = obj_detect_dots.size();

  if(obj_detect_dotsNum < min_detect_dotsNum)
  {
      ROS_WARN("min dots failed");
      return false;
  }

  std::sort(obj_detect_dots.begin(), obj_detect_dots.end(), [](const LASER_DOT_Typedef & dot_1,const LASER_DOT_Typedef & dot_2) { return dot_1.yaw < dot_2.yaw; });


  bool IfPreInRegion = false;

  REGION_Typedef region;

  for(int i = 0; i < obj_detect_dotsNum; i++)
  {
      bool IfInRegion;
      if(i == 0)
      {
        IfInRegion = iflaserDotsInRegion(obj_detect_dots[0],obj_detect_dots[obj_detect_dotsNum - 1]);
      }
      else
      {
        IfInRegion = iflaserDotsInRegion(obj_detect_dots[i],obj_detect_dots[i- 1]);
      }
      if(IfInRegion == true)
      {
          if(IfPreInRegion == false)
          {
              region.count = 0;
              region.Min_Yaw_Index = i;
          }
          region.Max_Yaw_Index = i;
          region.count++;
      }
      else if(region.count >= min_dots_in_region)
      {
          obj_regions.push_back(region);
      }
      IfPreInRegion = IfInRegion;
  }

  if(obj_regions.empty())
  {
      ROS_WARN("min dots in region failed");
      return false;
  }

  for(int i = 0;i < obj_regions.size(); i++)
  {
      int MinYawIndex = obj_regions[i].Min_Yaw_Index;
      int MaxYawIndex = obj_regions[i].Max_Yaw_Index;

      //printf("MinYawIndex %d,MaxYawIndex %d",MinYawIndex,MaxYawIndex);

      double FirstX = obj_detect_dots[MinYawIndex].x;
      double FirstY = obj_detect_dots[MinYawIndex].y;

      double SecondX = obj_detect_dots[MaxYawIndex].x;
      double Secondy = obj_detect_dots[MaxYawIndex].y;

      double length = sqrt((FirstX - SecondX) * (FirstX - SecondX) + (FirstY - Secondy) * (FirstY - Secondy));

      if(length < 0.5)
      {
         UTILITY_REGION_Typedef utility_region;
         utility_region.count = obj_regions[i].count;
         utility_region.MeanX = (FirstX + SecondX)/2.0;
         utility_region.MeanY = (FirstY + Secondy)/2.0;
         utility_obj_regions.push_back(utility_region);
      }
  }
  int utility_regionsNum = utility_obj_regions.size();
  ROS_WARN("utility_regionsNum %d",utility_regionsNum);
  if(utility_regionsNum == 0)
  {
      ROS_WARN("region length failed");
      return false;
  }
  std::sort(utility_obj_regions.begin(), utility_obj_regions.end(), [](const UTILITY_REGION_Typedef & region_1,const UTILITY_REGION_Typedef & region_2) { return region_1.count < region_2.count; });

  PublishPoint = utility_obj_regions[utility_regionsNum/2];

//  ROS_WARN("rival meanX %f, meanY %f",PublishPoint.MeanX,PublishPoint.MeanY);
  return true;
}

bool laser_detect::iflaserDotsInRegion(const LASER_DOT_Typedef& curDots,const LASER_DOT_Typedef& preDots)
{
    if(fabs(preDots.x - curDots.x) < max_two_dots_interval && fabs(preDots.y - curDots.y) < max_two_dots_interval)
    {
        return true;
    }
    return false;
}



void laser_detect::laserReceived(const sensor_msgs::LaserScanConstPtr& laser_scan)
{
//   boost::recursive_mutex::scoped_lock lr(configuration_mutex_);
   boost::mutex::scoped_lock lock(_mutex);
   ROS_WARN("detect    ");

  tf::Stamped<tf::Pose> laser_pose;

  if(!IfStaticMapRec)
  {
     ROS_WARN("STATIC_MAP NOT RECEIVE");
     return;
  }

  tf::Stamped<tf::Pose> ident (tf::Transform(tf::createIdentityQuaternion(),
                                            tf::Vector3(0,0,0)),laser_scan->header.stamp, laser_frame_id_);
  try
  {
    this->tf_->transformPose(global_frame_id_,ident,laser_pose);
  }
  catch(tf::TransformException& e)
  {
    ROS_ERROR("Couldn't transform from %s to %s, "
              "even though the message notifier is in use",
              laser_scan->header.frame_id.c_str(),
              global_frame_id_.c_str());
    return;
  }

    ROS_WARN("Received laser's pose wrt map: %.3f %.3f",
                laser_pose.getOrigin().x(),
                laser_pose.getOrigin().y());

    int range_count = laser_scan->ranges.size();

    double angle_min = laser_scan->angle_min;
    double angle_increment = laser_scan->angle_increment;

    // wrapping angle to [-pi .. pi]
    angle_increment = fmod(angle_increment + 5*M_PI, 2*M_PI) - M_PI;

    ROS_WARN("Lasers angles in base frame: min: %.3f inc: %.3f", angle_min, angle_increment);

    laser_scan_range = new double[range_count][2];

    for(int i = 0;i < range_count; i++)
    {
      laser_scan_range[i][0] = laser_scan->ranges[i];
      laser_scan_range[i][1] = angle_min + (i * angle_increment);
    }

    if(laser_scan_solve(range_count,laser_pose,PublishPoint))
    {
        laser_latest_detect_tf_ = tf::Transform(tf::Quaternion(0, 0, 0, 1), tf::Vector3(PublishPoint.MeanX, PublishPoint.MeanY, 0));
        laser_detect_tf_->sendTransform(tf::StampedTransform(laser_latest_detect_tf_,ros::Time::now(),global_frame_id_,detect_frame_id_));
    }
    delete laser_scan_range;
}



laser_detect::~laser_detect()
{
    delete tf_;
    delete laser_scan_sub_;
    delete laser_scan_filter_;

    delete laser_detect_tf_;

    //delete marker_pub;

    map_free( static_map_ );
}


void laser_detect::requesStatictMap()
{
    const float max_occ_dist = 8.0;
   // boost::recursive_mutex::scoped_lock ml(configuration_mutex_);
    // get map via RPC
    boost::mutex::scoped_lock lock(_mutex);
    nav_msgs::GetMap::Request  req;
    nav_msgs::GetMap::Response resp;
    ROS_INFO("Requesting the map...");
    while(!ros::service::call("static_map",req, resp))
    {
      ROS_WARN("Request for map failed; trying again...");
      ros::Duration MapDuration(0.5);
      MapDuration.sleep();
    }
    if( static_map_ != NULL )
    {
        map_free( static_map_ );
        static_map_ = NULL;
    }
    static_map_ = convertMap(resp.map);
    map_update_cspace(this->static_map_, max_occ_dist);
    IfStaticMapRec = true;
    ROS_WARN("maped...");
}

map_t* laser_detect::convertMap(const nav_msgs::OccupancyGrid& map_msg )
{
  map_t* map = map_alloc();
  ROS_ASSERT(map);

  map->size_x = map_msg.info.width;
  map->size_y = map_msg.info.height;
  map->scale = map_msg.info.resolution;
  map->origin_x = map_msg.info.origin.position.x + (map->size_x / 2) * map->scale;
  map->origin_y = map_msg.info.origin.position.y + (map->size_y / 2) * map->scale;
  // Convert to player format
  map->cells = (map_cell_t*)malloc(sizeof(map_cell_t)*map->size_x*map->size_y);
  ROS_ASSERT(map->cells);
  for(int i=0;i<map->size_x * map->size_y;i++)
  {
    if(map_msg.data[i] == 0)
      map->cells[i].occ_state = -1;
    else if(map_msg.data[i] == 100)
      map->cells[i].occ_state = +1;
    else
      map->cells[i].occ_state = 0;
  }

  return map;
}
