#include <pluginlib/class_list_macros.h>
#include "my_global_planner/global_planner.h"


 //register this planner as a BaseGlobalPlanner plugin
 PLUGINLIB_EXPORT_CLASS(my_global_planner::MyGlobalPlanner, nav_core::BaseGlobalPlanner)

 using namespace std;

 //Default Constructor
 namespace my_global_planner {

 MyGlobalPlanner::MyGlobalPlanner()
 :costmap_(NULL), IsInitialized(false)
 {

 }

 MyGlobalPlanner::~MyGlobalPlanner()
 {
   delete myPlanner;
 }




 void MyGlobalPlanner::initialize(std::string name, costmap_2d::Costmap2DROS* costmap_ros)
 {
   ROS_WARN("initialize my global planner");
   if(!IsInitialized)
   {


     ros::NodeHandle private_nh("~/" + name);
     private_nh.param("heuristic_weight", heuristic_weight_, 0.0f);


     costmap_ = costmap_ros->getCostmap();
     frame_id_ = costmap_ros->getGlobalFrameID();
     CostmapIndexArray = costmap_->getCharMap();

     myPlanner = new DijkPlanner(costmap_);
     pOrientationCal = new OrientationFilter(INTERPOLATE);

     private_nh.param("use_gradient_construct",myPlanner->use_gradient_construct_,true);
     //get the tf prefix
     ros::NodeHandle prefix_nh;
     tf_prefix_ = tf::getPrefixParam(prefix_nh);

     plan_pub_ = private_nh.advertise<nav_msgs::Path>("plan", 1);
     potential_pub_ = private_nh.advertise<nav_msgs::OccupancyGrid>("potential", 1);

     IsInitialized = true;
     ROS_INFO("My own global planner initialized!");
   }
   else
   {
     ROS_WARN("The global planner has been initialized!");
   }

 }

 bool MyGlobalPlanner::makePlan(const geometry_msgs::PoseStamped& start, const geometry_msgs::PoseStamped& goal,std::vector<geometry_msgs::PoseStamped>& globalplan )
 {
   //ROS_WARN("start make a plan");
   if(!IsInitialized)
   {
     ROS_ERROR("This planner need to be IsInitialized first");
     return false;
   }

    boost::mutex::scoped_lock lock(_mutex);

//    ros::NodeHandle n;
    std::string global_frame = frame_id_;

    //until tf can handle transforming things that are way in the past... we'll require the goal to be in our global frame
    if (tf::resolve(tf_prefix_, goal.header.frame_id) != tf::resolve(tf_prefix_, global_frame)) {
        ROS_ERROR(
                "The goal pose passed to this planner must be in the %s frame.  It is instead in the %s frame.", tf::resolve(tf_prefix_, global_frame).c_str(), tf::resolve(tf_prefix_, goal.header.frame_id).c_str());
        return false;
    }

    if (tf::resolve(tf_prefix_, start.header.frame_id) != tf::resolve(tf_prefix_, global_frame)) {
        ROS_ERROR(
                "The start pose passed to this planner must be in the %s frame.  It is instead in the %s frame.", tf::resolve(tf_prefix_, global_frame).c_str(), tf::resolve(tf_prefix_, start.header.frame_id).c_str());
        return false;
    }

    double WorldStart_X = start.pose.position.x;
    double WorldStart_Y = start.pose.position.y;
    double WorldGoal_X  = goal.pose.position.x;
    double WorldGoal_Y  = goal.pose.position.y;

    double MapNodeStart_X,MapNodeStart_Y,MapNodeGoal_X,MapNodeGoal_Y;

    //cost Map
    unsigned int CostmapStart_X, CostmapStart_Y, CostmapGoal_X, CostmapGoal_Y;
    if (!costmap_->worldToMap(WorldStart_X, WorldStart_Y, CostmapStart_X, CostmapStart_Y))
    {
         ROS_WARN("The robot's start position is off the global costmap. Planning will always fail, are you sure the robot has been properly localized?");
         return false;
    }

    if (!costmap_->worldToMap(WorldGoal_X, WorldGoal_Y, CostmapGoal_X, CostmapGoal_Y))
    {
         ROS_WARN("The goal sent to the global planner is off the global costmap. Planning will always fail to this goal.");
         return false;
    }
    //clear the starting cell within the costmap because we know it can't be an obstacle
    //set the associated costs in the cost map to be free
    costmap_->setCost(CostmapStart_X, CostmapStart_Y, costmap_2d::FREE_SPACE);


    //cost Map end


    worldToMap(WorldStart_X, WorldStart_Y, MapNodeStart_X, MapNodeStart_Y);
    worldToMap(WorldGoal_X,  WorldGoal_Y,  MapNodeGoal_X, MapNodeGoal_Y);

//    ROS_WARN("CostmapStart_X  %d  CostmapStart_Y  %d  CostmapGoal_X  %d  CostmapGoal_Y %d"
//              ,CostmapStart_X,CostmapStart_Y,CostmapGoal_X,CostmapGoal_Y);
    //plan_alogrithim
    if(myPlanner->global_path_seek(MapNodeStart_X, MapNodeStart_Y, MapNodeGoal_X, MapNodeGoal_Y,path))
    {
      if(PathToGlobalPlan(path,globalplan))
      {
        pOrientationCal->processPath(start,globalplan);
        publishPotential(myPlanner->Node_cost_dist);
        publishPlan(globalplan);
        return true;
      }
      ROS_ERROR("path empty");
    }
    ROS_ERROR("NO PATH");
    return false;
 }

bool MyGlobalPlanner::PathToGlobalPlan(std::vector<std::pair<float, float> >& path,std::vector<geometry_msgs::PoseStamped>& globalplan)
{
    std::string global_frame = frame_id_;

    //clear the plan
    globalplan.clear();

    ros::Time plan_time = ros::Time::now();
    for (int i = path.size() -1; i >= 0; i--)
    {
        std::pair<float, float> point = path[i];
        //convert the plan to world coordinates
        double world_x, world_y;
        mapToWorld(point.first, point.second, world_x, world_y);

        geometry_msgs::PoseStamped pose;
        pose.header.stamp = plan_time;
        pose.header.frame_id = global_frame;
        pose.pose.position.x = world_x;
        pose.pose.position.y = world_y;
        pose.pose.position.z = 0.0;
        pose.pose.orientation.x = 0.0;
        pose.pose.orientation.y = 0.0;
        pose.pose.orientation.z = 0.0;
        pose.pose.orientation.w = 1.0;
        globalplan.push_back(pose);
    }
    return !globalplan.empty();
}


 void MyGlobalPlanner::publishPotential(float const* SeekRegion)
 {
     int nx = costmap_->getSizeInCellsX(), ny = costmap_->getSizeInCellsY();
     double resolution = costmap_->getResolution();
     nav_msgs::OccupancyGrid grid;
     // Publish Whole Grid
     grid.header.frame_id = frame_id_;
     grid.header.stamp = ros::Time::now();
     grid.info.resolution = resolution;

     grid.info.width = nx;
     grid.info.height = ny;

     double wx, wy;
     costmap_->mapToWorld(0, 0, wx, wy);
     grid.info.origin.position.x = wx - resolution / 2;
     grid.info.origin.position.y = wy - resolution / 2;
     grid.info.origin.position.z = 0.0;
     grid.info.origin.orientation.w = 1.0;

     grid.data.resize(nx * ny);

     float max = 0.0;
     for (unsigned int i = 0; i < grid.data.size(); i++)
     {
         float potential = SeekRegion[i];
         if (potential < FLT_MAX)
         {
             if (potential > max)
             {
                 max = potential;
             }
         }
     }

     for (unsigned int i = 0; i < grid.data.size(); i++)
     {
       float potential = SeekRegion[i];
         if (potential >= FLT_MAX)
         {
             grid.data[i] = -1;
         }
         else
         {
            grid.data[i] = potential * 100 / max;
         }
     }
     potential_pub_.publish(grid);
 }

 void MyGlobalPlanner::publishPlan(const std::vector<geometry_msgs::PoseStamped>& plan)
 {
     //create a message for the plan
     nav_msgs::Path gui_path;
     gui_path.poses.resize(plan.size());

     gui_path.header.frame_id = frame_id_;
     gui_path.header.stamp = ros::Time::now();

     // Extract the plan in world co-ordinates, we assume the path is all in the same frame
     for (unsigned int i = 0; i < plan.size(); i++)
     {
         gui_path.poses[i] = plan[i];
     }

     plan_pub_.publish(gui_path);
 }

 };
